from skorch.llm.classifier import FewShotClassifier, ZeroShotClassifier
from skorch.llm.prompts import DEFAULT_PROMPT_FEW_SHOT, DEFAULT_PROMPT_ZERO_SHOT


__all__ = [
    'DEFAULT_PROMPT_FEW_SHOT',
    'DEFAULT_PROMPT_ZERO_SHOT',
    'FewShotClassifier',
    'ZeroShotClassifier',
]
